# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Twin-beam squeezer inverse-problem solver
=========================================
This module solves the *inverse* problem: given a joint photon number distribution find the best
parameters describing the different quantum states in a twin-beam producing it.

The ideas behind this module borrow heavily on the work of Burenkok et al. in

Full statistical mode reconstruction of a light field via a photon-number-resolved measurement
I. A. Burenkov, A. K. Sharma, T. Gerrits, G. Harder, T. J. Bartley, C. Silberhorn, E. A. Goldschmidt, and S. V. Polyakov
Phys. Rev. A 95, 053806 (2017)
"""

import numpy as np
from lmfit import Minimizer, Parameters
from sqtom.forward_solver import twinbeam_pmf


def two_schmidt_mode_guess(jpd_data, sq_label='sq_', noise_label='noise'):
    """Given a two mode histogram, this function generates a "physically" motivated guess for the loss, Schmidt occupations
    and dark counts parameters.

    This model is sensible only if the average g2 of signal and idler is above 1.5.

    Args:
        jpd_data (array): rectangular array with the probability mass functions of the photon events
        sq_label (string): label for the squeezing parameters.
        noise_label (string): label for the noise parameters.

    Returns:
        dict: dictionary containing a set of "reasonable" model parameters
    """
    res = marginal_calcs_2d(jpd_data)
    g2avg = np.max([0.5 * (res["g2_s"] + res["g2_i"]), 1.5])
    Nbar = 1.0 / (res["g11"] - g2avg)
    n0 = 0.5 * (1 + np.sqrt(2 * g2avg - 3)) * Nbar
    n1 = 0.5 * (1 - np.sqrt(2 * g2avg - 3)) * Nbar
    etas = res["n_s"] / Nbar
    etai = res["n_i"] / Nbar
    noise = np.abs(res["g2_s"] - res["g2_i"]) * Nbar
    return {
        "eta_s": etas,
        "eta_i": etai,
        sq_label + "0": n0,
        sq_label + "1": n1,
        noise_label + "_s": noise * etas,
        noise_label + "_i": noise * etai,
        "n_modes": 2,
    }


def marginal_calcs_2d(jpd_data, as_dict=True):
    """Given a two dimensional array of probabilities it calculates
    the mean photon numbers, their g2's and g11.

    It returns these values as a dictionary or as an array.

    Args:
        jpd_data (array): probability mass function of the photon events
        as_dict (boolean): whether to return the results as a dictionary

    Returns:
        dict or array: values of the mean photons number for signal and idlers, their corresponding g2 and their g11
    """
    inta, intb = jpd_data.shape
    na = np.arange(inta)
    nb = np.arange(intb)
    ns = np.sum(jpd_data, axis=1) @ na
    ni = np.sum(jpd_data, axis=0) @ nb
    ns2 = np.sum(jpd_data, axis=1) @ (na ** 2)
    ni2 = np.sum(jpd_data, axis=0) @ (nb ** 2)
    g2s = (ns2 - ns) / ns ** 2
    g2i = (ni2 - ni) / ni ** 2
    g11 = (na @ jpd_data @ nb) / (ns * ni)
    if as_dict is True:
        return {
            "n_s": ns,
            "n_i": ni,
            "g11": g11,
            "g2_s": g2s,
            "g2_i": g2i,
        }
    return np.array([ns, ni, g11, g2s, g2i])


def gen_hist_2d(beam1, beam2):
    """Calculate the joint probability mass function of events.

    Args:
        beam1 (array): 1D events array containing the raw click events of first beam
        beam2 (array): 1D events array containing the raw click events of second beam

    Returns:
        array: probability mass function of the click patterns for the two beams
    """
    nx = np.max(beam1)
    ny = np.max(beam2)
    xedges = np.arange(nx + 2)
    yedges = np.arange(ny + 2)
    mass_fun, _, _ = np.histogram2d(beam1, beam2, bins=(xedges, yedges), normed=True)
    return mass_fun


def fit_2d(
    pd_data, guess, do_not_vary=[], method="leastsq", cutoff=50, sq_label='sq_', noise_label='noise'
):
    """Returns a model fit from the parameter guess and the data

    Args:
        pd_data (array): one dimensional array of the probability distribution of the data
        guess (dict): dictionary with the guesses for the different parameters
        method (string): method to be used by the optimizer
        do_not_vary (list): list of variables that should be held constant during optimization
        cutoff (int): internal cutoff
        sq_label (string): label for the squeezing parameters.
        noise_label (string): label for the noise parameters.

    Returns:
        Object: object containing the optimized parameter and several goodness-of-fit statistics
    """
    pars_model = Parameters()
    n_modes = guess["n_modes"]
    pars_model.add("n_modes", value=n_modes, vary=False)
    for i in range(n_modes):
        pars_model.add(sq_label + str(i), value=guess["sq_" + str(i)], min=0.0)

    if "eta_s" in do_not_vary:
        pars_model.add("eta_s", value=guess["eta_s"], vary=False)
    else:
        pars_model.add("eta_s", value=guess["eta_s"], min=0.0, max=1.0)

    if "eta_i" in do_not_vary:
        pars_model.add("eta_i", value=guess["eta_i"], vary=False)
    else:
        pars_model.add("eta_i", value=guess["eta_i"], min=0.0, max=1.0)

    if noise_label + "_s" in do_not_vary:
        pars_model.add(noise_label + "_s", value=guess[noise_label + "_s"], vary=False)
    else:
        pars_model.add(noise_label + "_s", value=guess[noise_label + "_s"], min=0.0)

    if noise_label + "_i" in do_not_vary:
        pars_model.add(noise_label + "_i", value=guess[noise_label + "_i"], vary=False)
    else:
        pars_model.add(noise_label + "_i", value=guess[noise_label + "_i"], min=0.0)

    def model_2d(params, jpd_data):
        (dim_s, dim_i) = pd_data.shape
        joint_pmf = twinbeam_pmf(params, cutoff=cutoff)[:dim_s, :dim_i]
        return joint_pmf - pd_data


    minner_model = Minimizer(model_2d, pars_model, fcn_args=([pd_data]))
    result_model = minner_model.minimize(method=method)

    return result_model



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Degenerate squeezer inverse-problem solver
==========================================
This module solves the *inverse* problem of given a photon number distribution find the best
of parameters describing the different quantum states in a single beam producing it.

The ideas behind this module borrow heavily on the work of Burenkok et al. in

Full statistical mode reconstruction of a light field via a photon-number-resolved measurement
I. A. Burenkov, A. K. Sharma, T. Gerrits, G. Harder, T. J. Bartley, C. Silberhorn, E. A. Goldschmidt, and S. V. Polyakov
Phys. Rev. A 95, 053806 (2017)
"""


import numpy as np
from lmfit import Minimizer, Parameters
from sqtom.forward_solver import degenerate_pmf


def marginal_calcs_1d(pd_data, as_dict=True):
    """ Given a one dimensional array of probabilities it calculates the mean photon number
    and the g2.

    Args:
        pd_data (array): probability mass function of the photon events
        as_dict (boolean): whether to return the results as a dictionary

    Returns:
        dict or array: values of the mean photons number the corresponding g2.
    """

    intn = pd_data.shape[0]
    n = np.arange(intn)
    nmean = pd_data @ n
    nmean2 = pd_data @ n ** 2
    g2 = (nmean2 - nmean) / nmean ** 2
    if as_dict:
        return {"n": nmean, "g2": g2}
    return np.array([nmean, g2])


def threshold_1d(ps, nmax):
    """ Thresholds a probability distribution by assigning events with nmax
    photons or more to the nmax bin.

    Args:
        ps (array): probability distribution
        nmax (int): threshold value

    Returns:
        array: thresholded probability distribution
    """
    thr = nmax - 1
    local_ps = np.copy(ps)
    local_ps[thr] = np.sum(local_ps[thr:])
    return local_ps[:nmax]


def fit_1d(
    pd_data, guess, do_not_vary=None, method="leastsq", threshold=False, cutoff=50, sq_label='sq_', noise_label='noise'
):
    """Takes as input the name of the model to fit to and the jpd of the data
    and returns the fitted model.

    Args:
        pd_data (array): one dimensional array of the probability distribution of the data
        guess (dict): dictionary with the guesses for the different parameters
        method (string): method to be used by the optimizer
        do_not_vary (list): list of variables that should be held constant during optimization
        threshold (boolean or int): whether to threshold the photon probbailitites at value threshold
        cutoff (int): internal cutoff
        sq_label (string): string label for the squeezing parameters
        noise_label (string): label for the noise parameters.

    Returns:
        Object: object containing the optimized parameter and several goodness-of-fit statistics
    """
    if do_not_vary is None:
        do_not_vary = []

    pars_model = Parameters()
    n_modes = guess["n_modes"]
    pars_model.add("n_modes", value=n_modes, vary=False)
    # Add the squeezing parameters
    for i in range(n_modes):
        pars_model.add(sq_label + str(i), value=guess["sq_" + str(i)], min=0.0)

    if "eta" in do_not_vary:
        pars_model.add("eta", value=guess["eta"], vary=False)
    else:
        pars_model.add("eta", value=guess["eta"], min=0.0, max=1.0)

    if noise_label in do_not_vary:
        pars_model.add(noise_label, value=guess[noise_label], vary=False)
    else:
        pars_model.add(noise_label, value=guess[noise_label], min=0.0)

    if threshold:
        def model_1d(params, pd_data):
            ndim = pd_data.shape[0]
            dpmf = degenerate_pmf(params, cutoff=cutoff)
            return threshold_1d(dpmf, ndim) - pd_data
    else:
        def model_1d(params, pd_data):
            ndim = pd_data.shape[0]
            return degenerate_pmf(params, cutoff=cutoff, sq_label=sq_label, noise_label=noise_label)[:ndim] - pd_data

    minner_model = Minimizer(model_1d, pars_model, fcn_args=([pd_data]))
    return minner_model.minimize(method=method)



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Twin-beam and degenerate forward-problem solver
===============================================
This module solves the *forward* problem of given a set of parameters describing different
quantum state in two beams producing the joint photon number distribution. This joint
photon number distribution can then be passed to lmfit to solve the inverse problem using
the Levenberg-Marquardt method.

The ideas behind this module borrow heavily on the work of Burenkok et al. in

Full statistical mode reconstruction of a light field via a photon-number-resolved measurement
I. A. Burenkov, A. K. Sharma, T. Gerrits, G. Harder, T. J. Bartley, C. Silberhorn, E. A. Goldschmidt, and S. V. Polyakov
Phys. Rev. A 95, 053806 (2017)
"""


import numpy as np
from scipy.stats import poisson, geom
from scipy.signal import convolve2d
from thewalrus.quantum import loss_mat, gen_single_mode_dist


def twinbeam_pmf(params, cutoff=50, sq_label='sq_', noise_label='noise'):
    r"""Contructs the joint probability mass function of a conjugate source.

    Args:
        params (dict): Parameter dictionary, with possible keys "noise_s", "noise_i" for the
        Poisson noise mean photon numbers, "eta_s", "eta_i" for the transmission of the twin_beams,
        "n_modes" describing the number of twin_beams and the parameters sq_0,..,sq_n where
        n = n_modes giving the mean photon numbers of the different twin_beams.
        cutoff (int): Fock cutoff.
        sq_label (string): label for the squeezing parameters.
        noise_label (string): label for the noise parameters.

    Returns:
        (array): `n\times n` matrix representing the joint probability mass function
    """
    if noise_label + "_s" in params:
        noise_s = float(params[noise_label + "_s"])
    else:
        noise_s = 0.0

    if noise_label + "_i" in params:
        noise_i = float(params[noise_label + "_i"])
    else:
        noise_i = 0.0

    if "eta_s" in params:
        eta_s = float(params["eta_s"])
    else:
        eta_s = 1.0

    if "eta_i" in params:
        eta_i = float(params["eta_i"])
    else:
        eta_i = 1.0

    # First convolve all the 1-d distributions.
    ns = poisson.pmf(np.arange(cutoff), noise_s)
    ni = poisson.pmf(np.arange(cutoff), noise_i)

    joint_pmf = np.outer(ns, ni)
    # Then convolve with the twin beam distributions if there are any.
    if "n_modes" in params:
        n_modes = int(params["n_modes"])
        sq = [float(params[sq_label + str(i)]) for i in range(n_modes)]
        loss_mat_ns = loss_mat(eta_s, cutoff).T
        loss_mat_ni = loss_mat(eta_i, cutoff)
        twin_pmf = np.zeros([cutoff, cutoff])
        twin_pmf[0, 0] = 1.0
        for nmean in sq:
            twin_pmf = convolve2d(
                twin_pmf,
                np.diag(geom.pmf(np.arange(1, cutoff + 1), 1 / (1.0 + nmean),)),
            )[0:cutoff, 0:cutoff]
        twin_pmf = loss_mat_ns @ twin_pmf @ loss_mat_ni
        joint_pmf = convolve2d(twin_pmf, joint_pmf)[:cutoff, :cutoff]

    return joint_pmf


def degenerate_pmf(params, cutoff=50, sq_label='sq_', noise_label='noise'):
    r"""Contructs the probability mass function of a degenerate squeezing source.

    Args:
        params (dict): Parameter dictionary, with possible keys "noise" for the
        Poisson noise mean photon number, "eta", for the loss transmission, "n_modes" 
        describing the number of squeezed modes and the parameters sq_0,..,sq_n where
        n = n_modes giving the mean photon numbers of the different squeezers.
        cutoff (int): Fock cutoff.
        sq_label (string): label for the squeezing parameters.
        noise_label (string): label for the noise parameters.

    Returns:
        (array): `n\times n` matrix representing the joint probability mass function
    """
    if noise_label in params:
        noise = float(params[noise_label])
    else:
        noise = 0.0

    if "eta" in params:
        eta = float(params["eta"])
    else:
        eta = 1.0

    ps = poisson.pmf(np.arange(cutoff), noise)

    if "n_modes" in params:
        n_modes = int(params["n_modes"])
        sq = [float(params[sq_label + str(i)]) for i in range(n_modes)]
        mat = loss_mat(float(eta), cutoff)
        for n_val in sq:
            ps = np.convolve(
                ps,
                gen_single_mode_dist(np.arcsinh(np.sqrt(n_val)), cutoff=cutoff) @ mat,
            )[:cutoff]

    return ps[:cutoff]



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Basic setup module"""
from setuptools import setup


requirements = [
    "numpy",
    "scipy",
    "numba",
    "lmfit",
    "pytest",
    "thewalrus",
]

setup(
    name="sqtom",
    version="0.1",
    description="squeezing mode tomography",
    url="https://github.com/XanaduAI/sqtom",
    author="Xanadu",
    author_email="nicolas@xanadu.ai",
    license="Apache",
    packages=["sqtom"],
    install_requires=requirements,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type='text/markdown',
)
